import * as fs from 'fs'
import * as path from 'path'
import {
  AuxVmDetails,
  Driver,
  Env,
  ExecResult,
  GetTaskSetupDataResult,
  ScoringResult,
  TaskSetupData,
  TeardownResult,
  VmImageBuilder,
} from './Driver'
import { createAuxVm } from './src/aws'

export class AuxVMPermissionsError extends Error {}

function getRequiredEnv(taskSetupData: TaskSetupData, env: Env): Env {
  const missingEnvironmentVariables = taskSetupData.requiredEnvironmentVariables.filter(key => !(key in env))
  if (missingEnvironmentVariables.length > 0) {
    throw new Error(
      `The following required environment variables are not set: ${missingEnvironmentVariables.join(', ')}`,
    )
  }

  return Object.fromEntries(
    Object.entries(env).filter(([key]) => taskSetupData.requiredEnvironmentVariables.includes(key)),
  )
}

let taskHelperCode: string | undefined
function getDefaultTaskHelperCode(): string {
  if (taskHelperCode == null) {
    taskHelperCode = readFromAncestor('./task-standard/drivers/taskhelper.py')
  }
  return taskHelperCode
}

export function readFromAncestor(relativePath: string): string {
  let currentDir = __dirname
  const root = path.parse(currentDir).root

  while (currentDir !== root) {
    const filePath = path.resolve(currentDir, relativePath)
    try {
      const data = fs.readFileSync(filePath, 'utf8')
      return data
    } catch {
      currentDir = path.dirname(currentDir)
    }
  }
  throw new Error(`File not found: ${relativePath}`)
}

export class DriverImpl extends Driver {
  static readonly taskSetupDataSeparator = 'SEP_MUfKWkpuVDn9E'
  private static readonly taskNotFoundIndicator = 'taskNotFound_FPW3SDMlvf9Kf'
  constructor(
    readonly taskFamilyName: string,
    readonly taskName: string,
    // dockerExec MUST be a function that calls `docker container exec` or `docker container run` to execute a command
    // on a Docker container. dockerExec MUST forward its user, workdir, and env arguments to the `docker container exec`
    // or `docker container run` command.
    readonly dockerExec: (args: {
      pythonCode: string
      args?: string[]
      user: string
      workdir: string
      env: Env
    }) => Promise<ExecResult>,
    readonly taskHelperCode: string = getDefaultTaskHelperCode(),
  ) {
    super(taskFamilyName, taskName)
  }

  override async getTaskSetupData(): Promise<GetTaskSetupDataResult> {
    const execResult = await this.runTaskHelper('setup')

    if (execResult.stdout.includes(DriverImpl.taskNotFoundIndicator)) {
      return { status: 'taskNotFound' }
    }

    if (execResult.exitStatus !== 0) {
      return { status: 'processFailed', execResult }
    }

    let json: any
    try {
      json = JSON.parse(execResult.stdout.split(DriverImpl.taskSetupDataSeparator)[1].trim())
    } catch (e) {
      return { status: 'parseFailed', message: `Failed to parse task setup data.\n${e}` }
    }
    const taskSetupData = TaskSetupData.safeParse(json)
    if (!taskSetupData.success) {
      const errorMessages =
        taskSetupData.error.errors
          .map((error: any, index: number) => `${index + 1}. '${error.message}' at ${error.path?.join('.')}`)
          .join('\n') ?? 'No error messages found.'
      const message = `Failed to parse task setup data.\nCheck the get_permissions, get_instructions, required_environment_variables, and get_aux_vm_spec methods to ensure they're returning valid values.\nErrors:\n${errorMessages}\nJSON: ${JSON.stringify(json, null, 2)}\n`
      return { status: 'parseFailed', message }
    }
    return { status: 'succeeded', taskSetupData: taskSetupData.data }
  }

  override async maybeCreateAuxVm(
    taskEnvironmentIdentifier: string,
    taskFamilyDirectory: string,
    taskSetupData: TaskSetupData,
    buildVmImage: VmImageBuilder,
  ): Promise<AuxVmDetails | null> {
    if (taskSetupData.auxVMSpec == null) {
      return null
    }

    if (taskSetupData.permissions.length === 0 || !taskSetupData.permissions.includes('full_internet')) {
      throw new AuxVMPermissionsError(
        'DriverImpl only supports creating aux VMs in task environments with full internet access. We plan to change this in the future.',
      )
    }

    return await createAuxVm(taskEnvironmentIdentifier, taskFamilyDirectory, taskSetupData.auxVMSpec, buildVmImage)
  }

  override async startTask(taskSetupData: TaskSetupData, env: Env): Promise<void> {
    await this.runTaskHelper('start', { taskSetupData, env })
  }

  override async teardown(taskSetupData: TaskSetupData, env: Env): Promise<TeardownResult> {
    const execResult = await this.runTaskHelper('teardown', { taskSetupData, env })

    if (execResult.exitStatus !== 0) {
      return { status: 'processFailed', execResult }
    }
    const lastLine = execResult.stdout.split('\n').at(-1)!.trim()

    if (lastLine === 'None') return { status: 'noTeardown' }

    return { status: 'teardownSucceeded' }
  }

  override async scoreTask(submission: string, taskSetupData: TaskSetupData, env: Env): Promise<ScoringResult> {
    const execResult = await this.runTaskHelper('score', { submission, taskSetupData, env })
    if (execResult.exitStatus !== 0) {
      return { status: 'processFailed', execResult }
    }

    const lastLine = execResult.stdout.split('\n').at(-1)!.trim()
    if (lastLine === 'None') return { status: 'noScore' }

    const score = parseFloat(lastLine)
    if (isNaN(score)) {
      return { status: 'scoreWasNaN', execResult }
    }

    return { status: 'scoringSucceeded', score }
  }

  async runTaskHelper(
    operation: 'setup' | 'start' | 'score' | 'teardown',
    opts: { submission?: string; taskSetupData?: TaskSetupData; env?: Env } = {},
  ) {
    const args = [this.taskFamilyName, this.taskName, operation]
    if (opts.submission) {
      args.push('--submission', opts.submission)
    }
    return await this.dockerExec({
      pythonCode: this.taskHelperCode,
      args,
      user: 'root',
      workdir: '/root',
      env: opts.env && opts.taskSetupData ? getRequiredEnv(opts.taskSetupData, opts.env) : {},
    })
  }
}