import { createHash } from 'crypto'
import { lstat, readFile, readdir } from 'fs/promises'
import hash from 'object-hash'
import { join } from 'path'
import { FileBuildStep, VMSpec } from '../../Driver'
import { getImageNameGenerationData } from './getImageNameGenerationData'
import { getPackerTemplate } from './getPackerTemplate'

async function hashFileOrDirectory(path: string): Promise<string[]> {
  let fileHashes: string[] = []

  async function recurse(currentPath: string) {
    const stats = await lstat(currentPath)

    if (stats.isDirectory()) {
      const entries = await readdir(currentPath)
      for (let entry of entries) {
        await recurse(join(currentPath, entry))
      }
      return
    }

    const content = await readFile(currentPath)
    fileHashes.push(createHash('sha256').update(content).update(stats.mode.toString()).digest('hex'))
  }

  await recurse(path)
  return fileHashes
}

export async function getAuxVmImageName(taskFamilyDirectory: string, vmSpec: VMSpec): Promise<string> {
  const imageNameGenerationData = await getImageNameGenerationData(vmSpec)

  const fileBuildSteps = imageNameGenerationData.buildSteps.filter(({ type }) => type === 'file') as FileBuildStep[]
  const pathsToCopySourceFileHashes = (
    await Promise.all(fileBuildSteps.map(({ source }) => hashFileOrDirectory(join(taskFamilyDirectory, source))))
  ).flat()

  const vmSpecHash = hash({
    imageNameGenerationData,
    pathsToCopySourceFileHashes,
    packerTemplate: await getPackerTemplate(taskFamilyDirectory, imageNameGenerationData.buildSteps),
  })
  return `metr-task-standard-aux-vm-image-${vmSpecHash}`
}
